

# ==============================================================================
#                               NOISY CE ALGORITHM
#									(vers. 1.0)
# ==============================================================================
#
# Input:
# S(x,u)       - objective function (x: variables; u: noise)
# rugen(n,x)   - random sampling function for the noise
# v0           - starting value of parameters
# domain       - vector of domains of the variables "x" of "S(x,u)". The j-th
#				 element of "domain" specifies whether x[j] is either real
#                (value "0") or positive (value "1") or in (0,1) (value "2").
#                Default is real for the whole vector
# N            - size of sample drawn from "f" distribution (see algorithm)
# rho          - rho parameter (see algorithm)
# indsmooth2   - index of elements of parameter vector having a varying
#                smoothing parameter computed according to equation (5.1),
#                pag. 189 (Rubinstein and Kroese, 2004).
# asmooth      - constant smoothing coefficient
# bsmooth      - beta parameter of varying smoothing coefficient - see
#                eq. (5.1), pag. 189, Rubinstein and Kroese, 2004 -
# qsmooth      - q parameter of varying smoothing coefficient - see eq. (5.1),
#                pag. 189, Rubinstein and Kroese, 2004 -
# pval		   - threshold of p-value of the Geweke's test which triggers the 
#				 stopping of the algorithm
# buff		   - initial minimum size of the sample used by the Geweke's test
# vbuff        - speed of growth of the sample used by the Geweke's test. Must
#				 be a value in (0,1)
# pgeweke	   - proportions of Geweke's test
# maxiter      - maximum number of iterations of the algorithm
# rf(n,par)    - random sampling function from "f" distribution (see algorithm)
# vest(x)      - estimator of "f" distribution parameters (see algorithm)
# outEV		   - if TRUE...
#
# Remark on functions in input:
#   All the functions can have multivariate input and output. Sequences of 
# random vectors are stored in (n x k) matrices, where k is the dimension of
# the vector and n the number of vectors. This standard holds for all pdfs and
# random generators. If the parameter spaces have dimension d, input parameters
# "par" and output of "vest" should be d-dimensional vectors.
# ==============================================================================


noisyCE<- function(S,rugen,v0,domain=rep(0,length(v0)/2),N,rho=0.1,
  indsmooth2=2*(1:(length(v0)/2)),asmooth=1,bsmooth=0.9,qsmooth=10,pval=0.1,buff=10,
  vbuff=0.5,pgeweke=c(0.3,0.4),maxiter=1000,rf=NULL,vest=NULL,outEV=TRUE)
{
  require(magrittr)
  require(coda)
  
  # Define the "ratio" function (where 0/0=1)
  ratio<- function(x,y) { out<- x/y; out[is.nan(out)]<- 1; out}
  
  # Define the default settings
  if (is.null(rf) | is.null(vest))
  {
  	# Define the default estimators
    vest<- function(x)
  	{
  	  out<- NULL
  	  for (i in length(domain):1)
  	  {
  	  	if (domain[i]==0) { x[,i] %>% {c(mean(.),sd(.))} %>% c(out) -> out }
  	  	if (domain[i]==1) { x[,i] %>% log %>% {c(mean(.),sd(.))} %>% c(out) -> out }
  	  	if (domain[i]==2) { x[,i] %>% {log((1+.)/(1-.))} %>% {c(mean(.),sd(.))} %>% c(out) -> out }
  	  }
  	  out
    }
    # Define the default random generator
    rf<- function(n,par)
  	{
  	  out<- NULL
  	  for (i in 1:length(domain))
  	  {
  	  	if (domain[i]==0) { out %<>% cbind(rnorm(n,par[2*i-1],par[2*i])) }
  	  	if (domain[i]==1) { out %<>% cbind(rlnorm(n,par[2*i-1],par[2*i])) }
  	  	if (domain[i]==2) { rnorm(n,par[2*i-1],par[2*i]) %>% {2*(1+exp(-.))^(-1)-1} %>% cbind(out,.) ->out }
  	  }
  	  unname(out)
    }
  }
  
  # Initialisations
  n<- 1
  v<- matrix(v0,1)
  gam<- NA
  hpval<- rep(NA,buff)
  winf<- 1
  converged<- -1
  
  while(converged<0 & n<=maxiter)
  {
    # Update the counter
    n %<>% add(1)
    # Simulate the N values of the independent variables
    X<- rf(N,v[n-1,])
    # Simulate the N values of the objective function S
    {function(x) { S(x,rugen(x)) }} %>% apply(X,1,.) -> vettH
    # Compute and store the new gamma parameter
    vettH %>% quantile(probs=1-rho,type=1) %>% c(gam,.) -> gam
    # Compute the vector of H function values and store the optimal estimates
    vettH %>% is_weakly_greater_than(gam[n]) %>%
      {which(.==TRUE)} %>% X[.,] %>% vest %>% rbind(v,.) -> v
    # Compute the smoothing parameters
    sm<- rep(asmooth,ncol(v))
    if (!is.null(indsmooth2)) { sm[indsmooth2]<- bsmooth*(1-1/(n-1))^qsmooth }
    # Smooth the parameters
    v[n,]<- sm*v[n,]+(1-sm)*v[n-1,]
    # Verify the convergence
    #cat(paste0('Estimates [n=',n,']: '),formatC(v[n,],digits=6,format='f'),'\n')
    if (n>buff)
    {
      gam[-(1:winf)] %>% as.mcmc %>% geweke.diag(frac1=pgeweke[1],frac2=pgeweke[2]) %>% use_series(z) %>%
        abs %>% pnorm(lower.tail=FALSE) %>% c(hpval,.) -> hpval
      #%T>% {cat(paste0('p-value   [n=',n,']: '),formatC(.,digits=6,format='f'),'\n')}
      hpval %>% extract(n) %>% is_greater_than(pval) %>% subtract(1) -> converged
      if (is.na(converged)) { converged<- 1 }
      winf %<>% add(1-vbuff)
    }
  }
  #
  if (outEV==TRUE)
  {
  	depo<- {v[,which(domain==1)*2-1]+v[,which(domain==1)*2]^2/2} %>% exp
  	v[,which(domain==1)*2]<- {2*v[,which(domain==1)*2-1]+v[,which(domain==1)*2]^2} %>% exp %>% multiply_by(exp(v[,which(domain==1)*2]^2)-1) %>% sqrt
  	v[,which(domain==1)*2-1]<- depo
  	if (2 %in% domain)
  	{
  		# To be tested!!! -- Not used in the simulations of the paper!!!
  		dfcorr<- function(x,mu=0,sigma=1) { 2/(1-x^2)*dnorm(log((1+x)/(1-x)),mu,sigma) }
  		momdfcorr<- function(r,mu=0,sigma=1) { {function(x) {x^r*dfcorr(x,mu,sigma)}} %>% integrate(-1,1) %>% use_series(value) }
  		v %>% nrow ->depo
		depo %>% {outer(1:.,which(domain==2)*2-1,FUN=function(x,y) { (y-1)*10+x })} %>% matrix(ncol=1) ->pos
  		v[,which(domain==2)*2-1]<- apply(pos,1,function(x) {momdfcorr(1,v[x],v[x+depo])}) %>% matrix(nrow=depo)
  		v[,which(domain==2)*2]<- apply(pos,1,function(x) {momdfcorr(2,v[x],v[x+depo])}) %>% matrix(nrow=depo)	
  	}
  }
  # Output
  out<- list(estimate=unname(v[n,]),hestimate=unname(v),hgam=unname(gam),hpval=unname(hpval),niter=n,code=unname(converged))
  out
}
# ==============================================================================





# ==============================================================================
#  Test procedure
# ==============================================================================

prv1<- function()
{
  require(magrittr)
  
  camp<- rnorm(100,3,2)
  S<- function(x,u) { dnorm(camp,x[1],x[2]) %>% log %>% sum %>% add(u) %>% return }
  rugen<- function(par) { rnorm(1,0,0.01) }
  out<- noisyCE(S,rugen,c(0,10,0,10),domain=0:1,1000,0.01,asmooth=0.3,bsmooth=0.99,qsmooth=10,
    pval=0.1,vbuff=0.5,pgeweke=c(0.3,0.4),buff=10,maxiter=1000)

  cat('==================================\nMax:',S(c(mean(camp),sqrt(0.99*var(camp))),0),'\n')
  cat('==================================\nMLE:\n')
  cat('mu:',mean(camp),'; sig:',sqrt(0.99*var(camp)),'\n')
  cat('==================================\nCE estimate:\n')
  cat('mu:',out$est[1],'; sig:',exp(out$est[3]-out$est[4]^2/2),'\n==================================\n')
  alarm()
  invisible(out)
}
# ==============================================================================
